from colorama import Fore
from datetime import datetime
from socket import gethostbyname, gethostname
from json import dumps


def log(message: str):
    print(f"{Fore.LIGHTBLUE_EX}[{datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")}]: {Fore.RESET}{message}")


def ip_str(ip: str, port: int | None = None) -> str: return f"{Fore.GREEN}{ip}{f":{port}" if port is not None else ""}{Fore.RESET}"


def get_ip() -> str: return gethostbyname(gethostname())


def decrypt_message(message: bytes) -> list[str] | None:
    if message == b'': return None

    message = message.decode("utf-8")
    messages = message.split("\x00")[:-1]

    return messages


def encrypt_message(message_data: dict | str) -> bytes:
    if isinstance(message_data, str):
        message_data = {
            "text": message_data,
            "timestamps": [(get_ip(), datetime.now().isoformat())],
            "type": "message"
        }

    return f"{dumps(message_data)}\x00".encode("utf-8")


__all__ = ["log", "ip_str", "get_ip", "decrypt_message", "encrypt_message"]



from json import dump, load
from os import makedirs
from os.path import join, exists
from typing import Optional as Op

from anpm import aliaFolder
from requests import get
from requests.exceptions import ConnectionError

ygo_folder = join(aliaFolder, "Yu-Gi-Oh")
makedirs(ygo_folder, exist_ok=True)


def get_dump_return(data_url, file_path, ignore_exist=False):
    def get_return():
        data_ = get(data_url).json()
        dump(data_, open(file_path, "w"), indent=2)
        return data_

    def load_return():
        return load(open(file_path))

    if ignore_exist and exists(file_path):
        data = load_return()

    else:
        try:
            data = get_return()
        except ConnectionError:
            data = load_return()

    return data


card_info: list[dict] = \
    get_dump_return("https://db.ygoprodeck.com/api/v7/cardinfo.php", join(ygo_folder, "cards.json"), True)["data"]
card_info_by_name = {c["name"]: c for c in card_info}

card_sets: dict = get_dump_return("https://db.ygoprodeck.com/api/v7/cardsets.php", join(ygo_folder, "sets.json"), True)
card_sets_by_name = {s["set_name"]: s for s in card_sets}


class Card:
    # noinspection SpellCheckingInspection
    def __init__(self, raw_dict: dict):
        self.rawDict: dict = raw_dict

        #

        self._id: int = raw_dict["id"]
        self._name: str = raw_dict["name"]
        self._type: str = raw_dict["type"]
        self._humanReadableCardType: str = raw_dict["humanReadableCardType"]
        self._frameType: str = raw_dict["frameType"]
        self._desc: str = raw_dict["desc"]
        self._race: str = raw_dict["race"]
        self._ygoprodeck_url: str = raw_dict["ygoprodeck_url"]
        self._card_images: list[dict[str]] = raw_dict["card_images"]
        self._card_prices: list = raw_dict["card_prices"]

        self._archetype: Op[str] = raw_dict.get("archetype")
        self._card_sets: Op[list] = raw_dict.get("card_sets")
        self._typeline: Op[list[str]] = raw_dict.get("typeline")
        self._atk: Op[int] = raw_dict.get("atk")
        self._defence: Op[int] = raw_dict.get("def")
        self._level: Op[int] = raw_dict.get("level")
        self._attribute: Op[str] = raw_dict.get("attribute")
        self._linkval: Op[int] = raw_dict.get("linkval")
        self._linkmarkers: Op[list] = raw_dict.get("linkmarkers")
        self._pend_desc: Op[str] = raw_dict.get("pend_desc")
        self._monster_desc: Op[str] = raw_dict.get("monster_desc")
        self._scale: Op[int] = raw_dict.get("scale")
        self._banlist_info: Op[dict] = raw_dict.get("banlist_info")

        #

        self.name: str = self._name
        self.id: int = self._id
        self.type: str = self._type

        #

        self.earliestReleaseDate: Op[str] = self.get_earliest_date()
        self.croppedImages: list[str] = [i["image_url_cropped"] for i in self._card_images]

    def __repr__(self):
        return f"{type(self).__name__}({self.name})"

    def get_earliest_date(self):
        sets = self._card_sets
        if not sets: return
        return card_sets_by_name[sets[0]["set_name"]]["tcg_date"]


cardDatabase: list[Card] = [Card(c) for c in card_info]


def get_unique_values():
    from time import time

    path = join(ygo_folder, "unique_values.json")

    if not exists(path):
        value_ranges = {}

        x = 0
        len_card_info = len(card_info)

        start_time = time()

        for c in cardDatabase:
            x += 1
            print(f"{x / len_card_info * 100:.2f}%")

            for k, v in c.__dict__.items():
                if not k.startswith("_") or k.startswith("__"): continue

                if k not in value_ranges.keys(): value_ranges[k] = []
                if v not in value_ranges[k]: value_ranges[k].append(v)

        end_time = time()
        elapsed_time = end_time - start_time

        print(f"Finished in {elapsed_time:.2f} seconds!")

        # noinspection SpellCheckingInspection
        dump({k: v for k, v in value_ranges.items() if k not in [
            "_desc",
            "_ygoprodeck_url",
            "_card_images",
            "_card_prices",
            "_card_sets",
            "_card_prices",
            "_pend_desc",
            "_monster_desc",
        ]}, open(path, "w"), indent=2)

    value_ranges = load(open(path))

    return value_ranges


__all__ = ["card_info", "card_sets", "cardDatabase", "get_unique_values"]

if __name__ == '__main__':
    get_unique_values()



from socket import *
from threading import Thread
from datetime import datetime
from typing import Any, Callable

from .Shared import *


class Server(socket):
    def __init__(self):
        super().__init__()

        self.running: bool = False
        self.activeClients: list[tuple[socket, cat]] = []

        self.fClientConnect: Callable[[socket, cat], None] = lambda cs, ca: print(f"New connection from {ca[0]}.")
        self.fClientDisconnect: Callable[[socket, cat], None] = lambda cs, ca: print(f"{ca[0]} disconnected.")
        self.fRecieveData: Callable[[socket, cat, dict[str, Any]], None] = lambda cs, ca, mdata: print(mdata)
        self.fServerClose: Callable[[], None] = lambda: print("Server closed.")
        self.fServerStart: Callable[[str, int], None] = lambda ip, port: print(f"Now listening on {ip}:{port}.")

    def start(self, ip: str, port: int, backlog: int = 1):
        super().__init__(AF_INET, SOCK_STREAM)
        self.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self.bind((ip, port))
        self.listen(backlog)
        self.running = True

        def accept_clients():
            while self.running:
                try:
                    Thread(target=self.handle_client, args=(*self.accept(),)).start()
                except OSError as e:
                    if e.errno != 10038: raise e
                    break

        Thread(target=accept_clients, daemon=True).start()

        self.fServerStart(ip, port)

    def wait(self):
        while self.running: ...

    def stop(self):
        self.running = False
        self.fServerClose()
        self.close()

    def handle_client(self, client_socket: socket, client_address: tuple[str, int]):
        csa = (client_socket, client_address)
        self.activeClients.append(csa)

        self.fClientConnect(*csa)

        while self.running:
            message_data = client_socket.recv(1024)
            if not message_data: break

            for mdata in decipher_message(message_data):
                mdata["recieveTime"] = datetime.now().isoformat()

                self.fRecieveData(*csa, mdata)

            self.stop()

        self.activeClients.remove(csa)

        self.fClientDisconnect(*csa)


__all__ = ["Server"]

if __name__ == '__main__':
    server = Server()
    server.start("localhost", 1234)
    server.wait()



from inspect import currentframe
from os import environ, makedirs, listdir, walk
from os.path import join, expanduser, isdir, isfile
from queue import Queue
from re import split, escape
from threading import Thread
from typing import Any

from .variables import aliaFolder


def cycle_list(list_: list | tuple):
    if len(list_) == 1: return list_
    return list_[1:] + list_[:1]


def human_readable_size(size):
    for unit in ['B', 'KB', 'MB', 'GB', 'TB', 'PB']:
        if size < 1024:
            return f"{size:.2f} {unit}"
        size /= 1024


def get_files(root_: str = expanduser("~")) -> str:
    to_check_ = [root_]

    while to_check_:
        folder_ = to_check_.pop(0)

        try:
            contents_ = listdir(folder_)
        except PermissionError:
            continue

        for file_ in contents_:
            item = join(folder_, file_)

            if isdir(item):
                to_check_.append(item)
            elif isfile(item):
                yield item


def list_files_in_dir(root: str = "C:\\"):
    """AI GENERATED"""

    def worker_thread_(directory_, files_in_dir_, output_queue_):
        for file_name_ in files_in_dir_:
            file_path_ = join(directory_, file_name_)
            output_queue_.put(file_path_)

    file_queue_ = Queue()
    active_threads_ = []

    for current_directory_, subdirectories_, file_names_ in walk(root):
        thread_ = Thread(
            target=worker_thread_,
            args=(current_directory_, file_names_, file_queue_)
        )
        active_threads_.append(thread_)
        thread_.start()

    for thread_ in active_threads_:
        thread_.join()

    while not file_queue_.empty():
        yield file_queue_.get()


def default(current_var, default_value): return default_value if current_var is None else current_var


def gen_context(back: int = 1) -> dict[str, Any]:
    frame = currentframe()

    for _ in range(back):
        frame = frame.f_back

    return frame.f_locals


def re_split(input_string: str, delimiters: list[str]) -> list[str]:
    return split(f"[{''.join(map(escape, delimiters))}]", input_string)


def absorb(func):
    """Wrapper"""

    # noinspection PyUnusedLocal
    def do(*args, **kwargs):
        func()

    return do


class DefaultClass:
    def __repr__(
            self) -> str: return f"{type(self).__name__}({", ".join([f"{k}: {repr(v) if not callable(v) else v.__name__}" for k, v in self.__dict__.items()])})"


environ["TF_ENABLE_ONEDNN_OPTS"] = "0"
environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "1"
environ["HF_HUB_DISABLE_SYMLINKS_WARNING"] = "1"

makedirs(aliaFolder, exist_ok=True)

__all__ = ["aliaFolder", "cycle_list", "human_readable_size", "get_files", "list_files_in_dir", "default",
           "gen_context", "re_split", "absorb", "DefaultClass"]

if __name__ == '__main__':
    # DEBUGGING
    # TODO

    class Example:
        def __getitem__(self, args: slice):
            print(args.start, args.stop, args.step)


    test = Example()
    print(test["start":"stop":"step"])



from os.path import join, expanduser, exists
from json import load

formats = {
    "text": ["txt"],
    "image": ["png", "webp", "jpg", "jpeg"],
    "video": ["mp4", "mov", "m4a"],
    "sound": ["ogg", "mp3"]
}

keys = load(open(c)) if exists(c := join(expanduser("~"), "Code", "keys.json")) else {}

aliaFolder = join(expanduser("~"), "AppData", "Local", "Alia")

__all__ = ["formats", "keys", "aliaFolder"]



from tkinter import Frame, Event, Label
from typing import Literal, Callable
from tkinter.messagebox import showinfo

from anpm.tk import protect_grid_size, Border
from anpm import cycle_list


class Grid(dict[tuple[int, int], "Grid.Tile"]):
    def __init__(self):
        self.size = (0, 0)
        super().__init__()
        self.selected = None
        self.teams = []
        self.team = None
        self.running = False
        self.on_death = lambda: None

        self.outer_frame = Frame()
        self.inner_frame = Frame(self.outer_frame)
        self.inner_frame.grid_propagate(False)
        self.inner_frame.pack(expand=True)

        def update_inner_grid_size():
            outer_width = self.outer_frame.winfo_width()
            outer_height = self.outer_frame.winfo_height()

            grid_width, grid_height = self.size

            ratio = grid_width / grid_height if grid_height != 0 else 1

            if outer_width / outer_height > ratio:
                width = outer_height * ratio
                height = outer_height
            else:
                width = outer_width
                height = outer_width / ratio

            self.inner_frame.configure(width=width, height=height)

            for w in self.inner_frame.winfo_children():
                if isinstance(w, Label):
                    w.update_idletasks()
                    w.configure(wraplength=w.winfo_width())

        self.outer_frame.bind("<Configure>", lambda _: update_inner_grid_size())
        self.force_resize = update_inner_grid_size

    def update_grid(self, size: tuple[int, int], tile_colours: list[str] | tuple[str, ...], teams: list[str], on_death: Callable[[], None]):
        self.size = size
        self.selected = None
        self.teams = teams
        self.team = teams[0]
        self.running = True
        self.on_death = on_death

        for pos, tile in list(self.items()):
            tile.destroy()
            del self[pos]

        width, height = size

        colour_index = 0

        for y in range(height):
            self.inner_frame.columnconfigure(y, weight=1)

            for x in range(width):
                self.inner_frame.rowconfigure(x, weight=1)

                pos = (x, height - y - 1)

                tile = self.Tile(self, pos, tile_colours[colour_index])
                tile.grid(row=y, column=x, sticky="nsew")

                self.update({pos: tile})

                colour_index = (colour_index + 1) % len(tile_colours)

                Border(self.inner_frame).grid(row=0, rows=height, column=x, sticky="nsw")
                if x == width - 1: Border(self.inner_frame).grid(row=0, rows=width, column=x + 1, sticky="nsw")

            Border(self.inner_frame).grid(row=y, column=0, columns=height, sticky="new")
            if y == height - 1: Border(self.inner_frame).grid(row=y + 1, column=0, columns=height, sticky="new")

            if width / 2 == round(width / 2):
                colour_index = (colour_index + 1) % len(tile_colours)

        self.force_resize()

    class Tile(Label):
        def __init__(self, grid: "Grid", pos: tuple[int, int], colour: str):
            self.colour = colour
            self.pos = pos
            self.grid_ = grid
            self.piece = None

            super().__init__(grid.inner_frame, background=colour)
            self.bind("<ButtonRelease-1>", self.clicked)
            self.bind("<Button-1>", lambda _: self.pre_click())

        def clicked(self, event: Event):
            selected = self.grid_.selected

            if self.grid_.inner_frame.winfo_containing(event.x_root, event.y_root) != self:
                self.grid_.process_click(selected, None)
                return

            self.grid_.selected = self.pos
            self.grid_.process_click(selected, self.pos)

        def reachable(self):
            if self.piece is None: return []

            reachable_tiles = set()
            move_types = self.piece.move_types

            for m in move_types:
                to_process = [(self.pos, 0)]

                while to_process:
                    current_tile, index = to_process.pop()

                    if index < m[1]:
                        for t in self.grid_[current_tile].adjacent(m[0]):
                            team1 = self.piece.team if self.piece is not None else None
                            team2 = self.grid_[t].piece.team if self.grid_[t].piece is not None else None

                            if team1 == team2: ...

                            else:
                                if t != self.pos and t not in reachable_tiles: reachable_tiles.add(t)
                                if team2 is None: to_process.append((t, index + 1))

            return reachable_tiles

        def adjacent(self, move_type: Literal["horizontal", "vertical"]):
            x, y = self.pos

            co_ords = []

            if move_type in ["horizontal-e", "spread"]: co_ords.extend([(x + 1, y)])
            if move_type in ["horizontal-w", "spread"]: co_ords.extend([(x - 1, y)])

            if move_type in ["vertical-n", "spread"]: co_ords.extend([(x, y + 1)])
            if move_type in ["vertical-s", "spread"]: co_ords.extend([(x, y - 1)])

            if move_type in ["diagonal-ne"]: co_ords.extend([(x + 1, y + 1)])
            if move_type in ["diagonal-nw"]: co_ords.extend([(x - 1, y + 1)])
            if move_type in ["diagonal-se"]: co_ords.extend([(x + 1, y - 1)])
            if move_type in ["diagonal-sw"]: co_ords.extend([(x - 1, y - 1)])

            if move_type in ["knight"]: co_ords.extend([
                (x + 1, y + 2),
                (x - 1, y + 2),
                (x + 1, y - 2),
                (x - 1, y - 2),

                (x + 2, y + 1),
                (x - 2, y + 1),
                (x + 2, y - 1),
                (x - 2, y - 1),
            ])

            return [tile for tile in co_ords if tile in list(self.grid_.keys())]

        def pre_click(self):
            if self.grid_.selected is None: return

            if self.pos in self.grid_[self.grid_.selected].reachable():
                self.configure(background="#2a9e2a")

            elif self.pos == self.grid_.selected:
                self.configure(background="#1874cc")

    class Piece:
        def __init__(self, name: str, move_types: list[
            tuple[str, int] | tuple[str, int, str]
        ], team: str, king: bool = False):
            self.move_types = move_types
            self.name = name
            self.team = team
            self.king = king

    def process_click(self, tile1: tuple[int, int] | None, tile2: tuple[int, int] | None):
        if self.team is None or not self.running:
            self.selected = None
            return

        elif tile1 == tile2: self.selected = None
        elif tile1 is None and self[tile2].piece is not None and self[tile2].piece.team != self.team: self.selected = None
        elif tile2 is None: ...
        elif tile1 is None and self[tile2].piece is None: self.selected = None
        elif tile1 is None: ...
        elif self[tile2].piece is not None and self[tile1].piece.team == self[tile2].piece.team: self.selected = tile2
        elif tile2 not in self[tile1].reachable(): self.selected = tile1

        else:
            self.selected = None
            tile_2_piece = self[tile2].piece
            self[tile2].piece = self[tile1].piece
            self[tile1].piece = None

            if tile_2_piece is not None and tile_2_piece.king:
                self.check_win(tile_2_piece.team)

            else:
                self.teams = cycle_list(self.teams)
                self.team = self.teams[0]

        if self.running:
            self.update_tiles(self.selected)
            self.check_win()

    def update_tiles(self, tile1: tuple[int, int] = None):
        for pos, tile in self.items():
            if tile.piece is not None and tile.piece.team == self.team:
                tile.configure(foreground="#000000", font="TkDefaultFont 8 bold")
            else: tile.configure(foreground="#000000", font="TkDefaultFont 8")

            if pos == self.selected: tile.configure(background="#1e90ff", foreground="#000000")

            elif tile1 is not None and self[tile1].piece is not None and pos in self[tile1].reachable():
                tile.configure(background="#32cd32")

            else: tile.configure(background=tile.colour)

            if tile.piece is not None: tile.configure(text=tile.piece.name)
            else: tile.configure(text="")

            protect_grid_size(self.inner_frame)

    def place(self, pos: tuple[int, int], piece: "Grid.Piece"):
        self[pos].piece = piece
        self.update_tiles()

    def check_win(self, disqualify_team: str | None = None):
        if disqualify_team in self.teams:
            """
            for tile in self.values():
                if tile.piece is not None and tile.piece.team == disqualify_team: tile.piece = None
            """
            self.teams.remove(disqualify_team)

        if len(self.teams) == 1:
            self.update_tiles(self.selected)
            showinfo(main.title(), f"{self.teams[0].capitalize()} team wins!")
            self.running = False

            self.on_death()


rook = ["horizontal-e", "horizontal-w", "vertical-n", "vertical-s"]
bishop = ["diagonal-ne", "diagonal-nw", "diagonal-se", "diagonal-sw"]

if __name__ == '__main__':
    from anpm.tk import NewTk

    main = NewTk("Grid test")
    main.set_rows(1), main.set_columns(1)

    grid_ = Grid()
    grid_.outer_frame.grid(row=0, column=0, sticky="nsew")

    def chess():
        grid_.update_grid((8, 8), ("#f0d9b5", "#b58863"), ["white", "black"], chess)

        grid_.place((0, 0), grid_.Piece("WHITE ROOK", [(i, 8) for i in rook], "white"))
        grid_.place((1, 0), grid_.Piece("WHITE KNIGHT", [("knight", 1)], "white"))
        grid_.place((2, 0), grid_.Piece("WHITE BISHOP", [(i, 8) for i in bishop], "white"))
        grid_.place((3, 0), grid_.Piece("WHITE QUEEN", [(i, 8) for i in bishop + rook], "white"))
        grid_.place((4, 0), grid_.Piece("WHITE KING", [(i, 1) for i in rook + bishop], "white", True))
        grid_.place((5, 0), grid_.Piece("WHITE BISHOP", [(i, 8) for i in bishop], "white"))
        grid_.place((6, 0), grid_.Piece("WHITE KNIGHT", [("knight", 1)], "white"))
        grid_.place((7, 0), grid_.Piece("WHITE ROOK", [(i, 8) for i in rook], "white"))

        grid_.place((0, 7), grid_.Piece("BLACK ROOK", [(i, 8) for i in rook], "black"))
        grid_.place((1, 7), grid_.Piece("BLACK KNIGHT", [("knight", 1)], "black"))
        grid_.place((2, 7), grid_.Piece("BLACK BISHOP", [(i, 8) for i in bishop], "black"))
        grid_.place((3, 7), grid_.Piece("BLACK QUEEN", [(i, 8) for i in bishop + rook], "black"))
        grid_.place((4, 7), grid_.Piece("BLACK KING", [(i, 1) for i in rook + bishop], "black", True))
        grid_.place((5, 7), grid_.Piece("BLACK BISHOP", [(i, 8) for i in bishop], "black"))
        grid_.place((6, 7), grid_.Piece("BLACK KNIGHT", [("knight", 1)], "black"))
        grid_.place((7, 7), grid_.Piece("BLACK ROOK", [(i, 8) for i in rook], "black"))

    chess()

    main.mainloop()



from datetime import datetime
from json import dumps, loads
from typing import Any
from socket import socket


def send_message(socket_: socket, message: str):
    message_data = {
        "message": message,
        "sendTime": datetime.now().isoformat()
    }
    socket_.sendall(f"{dumps(message_data)}\x00".encode("utf-8"))


def decipher_message(message_data: str | bytes) -> list[dict[str, Any]]:
    if isinstance(message_data, bytes): message_data = message_data.decode("utf-8")
    return [loads(m) for m in message_data.split("\x00")[:-1]]


client_address_tuple = tuple[str, int]
cat = client_address_tuple

__all__ = ["send_message", "decipher_message", "cat"]



from socket import *

from .Shared import *


class Client(socket):
    def __init__(self):
        super().__init__()
        self.connected: bool = False

    def join(self, ip, port):
        super().__init__(AF_INET, SOCK_STREAM)
        self.connect((ip, port))

        self.connected = True

    def leave(self):
        self.connected = False
        self.shutdown(SHUT_RDWR)
        self.close()


__all__ = ["Client"]

if __name__ == '__main__':
    client = Client()
    client.join("localhost", 1234)

    send_message(client, "Hello World!")

    client.leave()



from socket import *

from threading import Thread
from datetime import datetime
from json import loads, dumps

try: from .shared import *
except ImportError: from shared import *


class Server(socket):
    def __init__(self):
        self.running: bool = False
        self.aliveThread: Thread | None = None
        self.activeClients: list[tuple[socket, tuple[str, int]]] = []

        super().__init__(AF_INET, SOCK_STREAM)

    def start(self, port: int, client_limit: int = 5):
        super().__init__(AF_INET, SOCK_STREAM)
        self.bind(("0.0.0.0", port))
        self.listen(client_limit)

        self.running = True

        self.aliveThread = Thread(target=self.while_alive)
        self.aliveThread.start()

        log(f"Started started on {ip_str(get_ip(), port)}...")

    def stop(self):
        self.running = False

        [cs.shutdown(SHUT_RDWR) for cs, _ in self.activeClients]

        self.activeClients.clear()

        self.close()

        log(f"Stopped server!")

    def while_alive(self):
        while self.running:
            try: Thread(target=self.handle_client, args=self.accept(), daemon=True).start()

            except OSError as e:
                if e.errno == 10038: ...
                else: raise e

    def handle_client(self, c_socket: socket, c_address: tuple[str, int]):
        self.activeClients.append((c_socket, c_address))

        log(f"{ip_str(*c_address)} connected!")

        while self.running:
            try:
                messages = c_socket.recv(1024)
                messages = decrypt_message(messages)

                if messages is None: break

                for message_data in messages:
                    message_data = loads(message_data)

                    log(f"{ip_str(*c_address)}: {message_data["text"]}")
                    message_data["timestamps"].append((get_ip(), datetime.now().isoformat()))

                    message_data = encrypt_message(message_data)

                    [cs.sendall(message_data) for cs, _ in self.activeClients]

            except ConnectionResetError: break

        self.activeClients.remove((c_socket, c_address))

        log(f"{ip_str(*c_address)} disconnected!")

        self.stop()


__all__ = ["Server"]

if __name__ == '__main__':
    server = Server()
    server.start(port=12345)



from socket import *

from threading import Thread
from datetime import datetime
from json import loads

try: from .shared import *
except ImportError: from shared import *


class Client(socket):
    def __init__(self, log_: bool = True):
        self.connected: bool = False
        self.aliveThread: Thread | None = None
        self.server_address: tuple[str, int] | None = None
        self.log = log_

        super().__init__(AF_INET, SOCK_STREAM)

        if self.log: log("Initiated client...")

    def join(self, ip: str, port: int, *, timeout_: float | None = None):
        self.settimeout(timeout_)

        self.connect((ip, port))

        self.server_address = (ip, port)
        self.connected = True
        self.aliveThread = Thread(target=self.while_alive)
        self.aliveThread.start()

        if self.log: log(f"Connected to {ip_str(ip, port)}!")

    def leave(self):
        self.connected = False

        self.shutdown(SHUT_RDWR)

        if self.log: log(f"Disconnected from {ip_str(*self.server_address)}!")

    def while_alive(self):
        while self.connected:
            messages = self.recv(1024)
            messages = decrypt_message(messages)

            if messages is None: break

            for message_data in messages:
                message_data = loads(message_data)

                message_data["timestamps"].append((get_ip(), datetime.now().isoformat()))

                self.handle_message(message_data)

    def send_message(self, message: str):
        self.send(encrypt_message(message))

    @staticmethod
    def handle_message(message_data: dict):
        log(f"{ip_str(message_data["timestamps"][0][0])}: {message_data["text"]}")


__all__ = ["Client"]

if __name__ == '__main__':
    client = Client()
    client.join("localhost", 12345)

    client.send_message("Hello World!")

    client.leave()



from socket import socket, AF_INET, SOCK_STREAM
from typing import Callable, Literal
from obswebsocket import obsws
from re import search
from requests import get, patch


# noinspection SpellCheckingInspection
class TwitchChat(socket):
    def __init__(self, username: str, token: str, channel: str, client_id: str):
        self.username = username
        self.token = f"oauth:{token}"
        self.channel = channel

        self.client_id = client_id
        self.api_token = token

        super().__init__(AF_INET, SOCK_STREAM)
        self.connect(("irc.chat.twitch.tv", 6667))
        self.send(f"PASS {self.token}\n".encode("utf-8"))
        self.send(f"NICK {self.username}\n".encode("utf-8"))
        self.send(f"JOIN #{self.channel}\n".encode("utf-8"))

    def listen_for_messages(self, called_function: Callable[[str, str], None | Literal["quit"]] = lambda username, message: print(f"{username}: {message.strip()}")):
        while True:
            response = self.recv(2048).decode("utf-8")

            if response.startswith("PING"):
                self.send("PONG :tmi.twitch.tv\n".encode("utf-8"))

            elif "PRIVMSG" in response:
                username = response.split("!")[0][1:]
                message = search(r"PRIVMSG[^:]*:(.*)", response).group(1)

                result = called_function(username, message)
                if result == "quit": break

    def send_message(self, message: str):
        self.send(f"PRIVMSG #{self.channel} :{message}\n".encode("utf-8"))
        print(f"Sent: {message}")

    def update_broadcast_info(self, title: str = None, game_id: str = None):
        url = "https://api.twitch.tv/helix/channels"
        headers = {
            "Authorization": f"Bearer {self.api_token}",
            "Client-Id": self.client_id,
        }
        data = {"broadcaster_id": self.get_broadcaster_id()}

        if title:
            data["title"] = title
        if game_id:
            data["game_id"] = game_id

        response = patch(url, headers=headers, json=data)
        if response.status_code == 204:
            print("Broadcast info updated successfully.")
        else:
            print(f"Failed to update broadcast info: {response.status_code} - {response.text}")

    def get_broadcaster_id(self):
        url = "https://api.twitch.tv/helix/users"
        headers = {
            "Authorization": f"Bearer {self.api_token}",
            "Client-Id": self.client_id,
        }
        params = {"login": self.channel}
        response = get(url, headers=headers, params=params)
        if response.status_code == 200:
            return response.json()["data"][0]["id"]
        else:
            raise Exception(f"Failed to get broadcaster ID: {response.status_code} - {response.text}")


class OBS(obsws):
    def __init__(self, host: str = "localhost", port: int = 4444, password: str = ""):
        super().__init__(host, port, password)
        self.connect()


__all__ = ["TwitchChat", "OBS"]



from setuptools import setup, find_packages

setup(
    name="anpm-public",
    version="0.1.0",
    description="A collection of scripts by me, Alia. THIS IS THE PUBLIC VERSION, VERSION  NUMBERS MAY BE INACURRATE.",
    author="Alia Normal",
    author_email="dan.driscoll@aussiebb.com",
    packages=find_packages(),
    install_requires=[],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.12",)



from platform import system
from tkinter import Tk, Misc, Widget, Frame, IntVar, Label, Toplevel
from tkinter.font import Font
from tkinter.ttk import Scrollbar as TtkScrollbar, Style
from typing import Literal, Callable, Any, Type

platform = system()


class NewTk(Tk):
    def __init__(self, title: str = "tk", resize: tuple[int | float, int | float] | None = (2, 2), center: bool = True,
                 **kwargs):
        if platform == "Windows":
            from ctypes import windll

            windll.shcore.SetProcessDpiAwareness(1)

        super().__init__(**kwargs)
        self.title(title)

        if center: self.center()
        if resize: self.resize(*resize)

        self.rows = IntVar(value=0)
        self.r = self.rows.get

        def update_rows(*_):
            for row in range(self.r()):
                self.grid_rowconfigure(row, weight=0)

        self.rows.trace_add("write", update_rows)

        self.columns = IntVar(value=0)
        self.c = self.columns.get

        def update_columns(*_):
            for column in range(self.c()):
                self.grid_columnconfigure(column, weight=0)

        self.columns.trace_add("write", update_columns)

        self.style = Style()

    def center(self):
        self.update_idletasks()

        screenwidth, screenheight = self.winfo_screenwidth(), self.winfo_screenheight()
        width, height = self.winfo_width(), self.winfo_height()

        self.geometry(f"+{(screenwidth - width) // 2}+{(screenheight - height) // 2}")

    def resize(self, width_factor: int | float, height_factor: int | float):
        self.update_idletasks()

        screenwidth, screenheight = self.winfo_screenwidth(), self.winfo_screenheight()
        new_width, new_height = screenwidth // width_factor, screenheight // height_factor

        x, y = self.winfo_x(), self.winfo_y()
        width, height = self.winfo_width(), self.winfo_height()
        new_x, new_y = x + (width - new_width) // 2, y + (height - new_height) // 2

        self.geometry(f"{int(new_width)}x{int(new_height)}+{int(new_x)}+{int(new_y)}")

    def set_rows(self, *row_weights: int):
        if len(row_weights) == 1 and row_weights[0] > 1:
            row_weights = [1] * row_weights[0]

        self.rows.set(len(row_weights))
        for row, weight in enumerate(row_weights):
            self.grid_rowconfigure(row, weight=weight)

    def set_columns(self, *column_weights: int):
        if len(column_weights) == 1 and column_weights[0] > 1:
            column_weights = [1] * column_weights[0]

        self.columns.set(len(column_weights))
        for column, weight in enumerate(column_weights):
            self.grid_columnconfigure(column, weight=weight)


def clear(window: Misc): [w.destroy() for w in window.winfo_children()]


def protect_grid_size(window: Misc):
    # noinspection PyArgumentList
    [widget.configure(width=1, height=1) for widget in window.winfo_children() if
     widget.cget("width") == 0 and len(widget.grid_info().get("sticky", "")) > 1]


class Border(Frame):
    def __init__(self, master: Misc | None = None, background: str = "#000000", width: int = 1,
                 height: int = 1): super().__init__(master, background=background, width=width, height=height)


class ScrollingFrame(Frame):
    from tkinter import Event

    def __init__(self, master: Misc | None = None, *_, scrollbar: bool = True, scrollbar_border: bool = True,
                 ttk_scrollbar: bool = False, **kwargs):
        from tkinter import Canvas, Scrollbar

        self.frame = Frame(master)
        self.frame.grid_propagate(False)
        self.frame.grid_columnconfigure(0, weight=1)
        self.frame.grid_rowconfigure(0, weight=1)

        self.canvas = Canvas(self.frame, highlightthickness=0)
        self.canvas.grid(row=0, column=0, sticky="nsew")

        if ttk_scrollbar:
            self.scrollbar = TtkScrollbar(self.frame, command=self.canvas.yview)
        else:
            self.scrollbar = Scrollbar(self.frame, command=self.canvas.yview)
        self.scrollbar_border = None

        if scrollbar:
            self.scrollbar.grid(row=0, column=1, sticky="ns")
            if scrollbar_border:
                self.scrollbar_border = Border(self.frame, "#a0a0a0")
                self.scrollbar_border.grid(row=0, column=0, sticky="nse")

        self.canvas.configure(yscrollcommand=self.scrollbar.set)

        self.canvas.bind("<Enter>", lambda _: self.canvas.bind_all("<MouseWheel>", self.mouse_scroll))
        self.canvas.bind("<Leave>", lambda _: self.canvas.unbind_all("<MouseWheel>"))

        super().__init__(self.canvas, **kwargs)
        self.id = self.canvas.create_window((0, 0), window=self, anchor="nw")

        self.canvas.bind("<Configure>", self.on_canvas_configure)
        self.bind("<Configure>", self.on_frame_configure)

    def grid(self, **kwargs):
        self.frame.grid(**kwargs)

    def pack(self, **kwargs):
        self.frame.pack(**kwargs)

    def place(self, **kwargs):
        self.frame.place(**kwargs)

    def mouse_scroll(self, event: Event):
        from tkinter import TclError

        try:
            current_scroll = self.canvas.yview()
            move = -0.045 if event.delta > 0 else 0.045
            new_scroll = current_scroll[0] + move
            self.canvas.yview_moveto(max(0.0, min(1.0, new_scroll)))
        except TclError:
            self.canvas.unbind_all("<MouseWheel>")

    def on_canvas_configure(self, event: Event):
        self.canvas.configure(scrollregion=self.canvas.bbox("all"))
        self.canvas.itemconfig(self.id, width=event.width)

    def on_frame_configure(self, event: Event):
        self.canvas.configure(background=self.cget("background"))
        self.on_canvas_configure(event)
        self.canvas.bind_all("<MouseWheel>", self.mouse_scroll)


class ToolTip(object):
    def __init__(self, widget: Widget, text: str | Callable[[], str], *_, delay: int | None = 0, x_offset: int = 30,
                 y_offset: int = 10,
                 follow: bool | Literal["once"] = False, **kwargs):
        from tkinter import Label, Toplevel, LEFT, SOLID

        self.widget = widget
        self.text = text
        self.delay = delay
        # noinspection SpellCheckingInspection
        self.kwargs = {
                          "justify": LEFT,
                          "background": "#ffffff",
                          "foreground": "#000000",
                          "font": None,
                          "relief": SOLID,
                          "borderwidth": 1,
                          "wraplength": 350
                      } | kwargs
        self.x_offset = x_offset
        self.y_offset = y_offset
        self.follow = follow

        self.id = None
        self.tw: Toplevel | None = None
        self.positioned = False
        self.label: Label | None = None

        self.enter_id: str = ""
        self.leave_id: str = ""
        self.buttonpress_id: str = ""
        self.motion_id: str = ""
        self.label_leave_id: str = ""
        self.bind_all()

    def enter(self):
        if self.tw is None:
            self.schedule()

    def leave(self):
        widget_under_pointer = self.widget.winfo_containing(*self.widget.winfo_pointerxy())
        if self.label is not None and (widget_under_pointer == self.label or widget_under_pointer == self.widget):
            return

        self.unschedule()
        self.hidetip()

    def schedule(self):
        self.unschedule()
        if self.delay is not None:
            self.id = self.widget.after(self.delay, self.showtip)

    def unschedule(self):
        if self.id and self.delay is not None:
            self.widget.after_cancel(self.id)
        self.id = None

    def showtip(self):
        from tkinter import Toplevel, Label
        if not self.tw:
            self.tw = Toplevel(self.widget)
            self.tw.wm_overrideredirect(True)

        if not self.positioned:
            self.update_tooltip_position()
            self.positioned = True

        self.label = Label(self.tw, text=self.text() if callable(self.text) else self.text, **self.kwargs)
        self.label.pack(ipadx=1)
        self.label.bind("<Leave>", lambda _: self.leave())

        self.label.IAmAToolTip = True

    def update_tooltip_position(self):
        if self.follow is True or self.follow == "once":
            x = self.widget.winfo_pointerx() + self.x_offset
            y = self.widget.winfo_pointery() + self.y_offset
        else:
            widget_x = self.widget.winfo_rootx() + self.x_offset
            widget_y = self.widget.winfo_rooty() + self.y_offset
            x, y = widget_x, widget_y

        self.tw.wm_geometry(f"+{x}+{y}")

    def on_motion(self):
        if self.tw and self.follow:
            self.update_tooltip_position()

    def bind_all(self):
        self.enter_id = self.widget.bind("<Enter>", lambda _: self.enter(), add=True)
        self.leave_id = self.widget.bind("<Leave>", lambda _: self.leave(), add=True)
        self.buttonpress_id = self.widget.bind("<ButtonPress>", lambda _: self.leave(), add=True)
        if self.follow is True:
            self.motion_id = self.widget.bind("<Motion>", lambda _: self.on_motion(), add=True)

    def destroy(self):
        self.widget.unbind("<Enter>", self.enter_id)
        self.widget.unbind("<Leave>", self.leave_id)
        self.widget.unbind("<ButtonPress>", self.buttonpress_id)
        if self.follow is True or self.follow == "once":
            self.widget.unbind("<Motion>", self.motion_id)
        self.leave()

    def hidetip(self):
        if self.tw:
            self.tw.destroy()
            self.tw = None
            self.positioned = False


class NewFrame(Frame):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.rows = IntVar(value=0)
        self.r = self.rows.get

        def update_rows(*_):
            for row in range(self.r()):
                self.grid_rowconfigure(row, weight=0)

        self.rows.trace_add("write", update_rows)

        self.columns = IntVar(value=0)
        self.c = self.columns.get

        def update_columns(*_):
            for column in range(self.c()):
                self.grid_columnconfigure(column, weight=0)

        self.columns.trace_add("write", update_columns)

    def set_rows(self, *row_weights: int):
        if len(row_weights) == 1 and row_weights[0] > 1:
            row_weights = [1] * row_weights[0]

        self.rows.set(len(row_weights))
        for row, weight in enumerate(row_weights):
            self.grid_rowconfigure(row, weight=weight)

    def set_columns(self, *column_weights: int):
        if len(column_weights) == 1 and column_weights[0] > 1:
            column_weights = [1] * column_weights[0]

        self.columns.set(len(column_weights))
        for column, weight in enumerate(column_weights):
            self.grid_columnconfigure(column, weight=weight)


def get_default_font():
    from tkinter.font import Font
    return Font(family="Segoe UI", size=9, weight="normal", slant="roman", underline=False, overstrike=False)


def is_hovered(widget: Widget):
    return widget.master.winfo_containing(*widget.master.winfo_pointerxy()) == widget


def validate_input(new_value: str, rule: Callable[[str], bool]) -> bool:
    return rule(new_value)


class NewFont(Font):
    def __init__(self):
        self.default = dict(family="Segoe UI", size=9, weight="normal", slant="roman", underline=False,
                            overstrike=False)
        super().__init__(**self.default)

    def reset(self):
        super().__init__(**self.default)
        return self


class NewLabel(Label):
    def __init__(self, master: Misc | None = None, cnf: dict[str, Any] | None = None, **kwargs):
        if cnf is None: cnf = {}

        self.frame = Frame(master)
        self.frame.pack_propagate(False)
        self.frame.grid_propagate(False)

        super().__init__(self.frame, cnf, **kwargs)
        super().pack(fill="both", expand=True)

    def pack(self, **kwargs): self.frame.pack(**kwargs)

    def grid(self, **kwargs): self.frame.grid(**kwargs)

    def destroy(self):
        super().destroy()
        self.frame.destroy()

    def pack_forget(self): self.frame.pack_forget()

    def place_forget(self): self.frame.place_forget()

    def grid_forget(self): self.frame.grid_forget()

    def super_pack(self, **kwargs): super().pack(**kwargs)

    def super_grid(self, **kwargs): super().grid(**kwargs)


class NewWidget(Widget):
    # def __init__(self, widget_type: Type[Widget], master: Misc | None = None, cnf: dict[str, Any] | None = None, *args, **kwargs):
    #     if cnf is None: cnf = {}

    # noinspection PyMissingConstructor
    def __init__(self, widget_type: Type[Widget], master: Misc | None = None, *args, **kwargs):
        self.frame = Frame(master)
        self.frame.pack_propagate(False)
        self.frame.grid_propagate(False)

        # self.widget = widget_type(self.frame, cnf, *args, **kwargs)
        self.widget = widget_type(self.frame, *args, **kwargs)
        self.widget.pack(fill="both", expand=True)

        self.widgetType = widget_type

    def pack(self, **kwargs): self.frame.pack(**kwargs)

    def grid(self, **kwargs): self.frame.grid(**kwargs)

    def destroy(self):
        self.widget.destroy()
        self.frame.destroy()

    def pack_forget(self): self.frame.pack_forget()

    def place_forget(self): self.frame.place_forget()

    def grid_forget(self): self.frame.grid_forget()

    def __getattr__(self, name):
        if hasattr(self.widget, name): return getattr(self.widget, name)
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    def super_pack(self, **kwargs): super().pack(**kwargs)

    def super_grid(self, **kwargs): super().grid(**kwargs)


def hovered_widget(widget: Widget) -> Widget | Tk | Misc:
    if isinstance(widget, Tk):
        return widget.winfo_containing(*widget.winfo_pointerxy())
    else:
        return widget.master.winfo_containing(*widget.master.winfo_pointerxy())


def generate_validate_command(window: Misc, rule: Callable[[str], bool]) -> tuple[str, str]:
    return window.register(lambda k: validate_input(k, rule)), "%P"


class LockedFrame(Frame):
    def __init__(
            self,
            master: Misc | None = None,
            ratio: tuple[int, int] | None = None,
            min_size: tuple[int, int] | None = None,
            max_size: tuple[int, int] | None = None,
            *args, **kwargs
    ):
        self.ratio = ratio
        self.min_width, self.min_height = min_size if min_size else (None, None)
        self.max_width, self.max_height = max_size if max_size else (None, None)

        self.frame = Frame(master)
        self.frame.pack_propagate(False)

        self.pack = self.frame.pack
        self.grid = self.frame.grid
        self.pack_forget = self.frame.pack_forget
        self.place_forget = self.frame.place_forget
        self.grid_forget = self.frame.grid_forget

        self.super_pack = super().pack
        self.super_grid = super().grid

        super().__init__(self.frame, *args, **kwargs)

        if any(i is not None for i in [ratio, min_size, max_size]):
            self.pack_propagate(False)
            self.grid_propagate(False)

        self.super_pack(expand=True)
        self.frame.bind("<Configure>", self._resize)
    def _resize(self, event):
        width, height = event.width, event.height

        if self.min_width and width < self.min_width:
            width = self.min_width

        if self.min_height and height < self.min_height:
            height = self.min_height

        if self.max_width and width > self.max_width:
            width = self.max_width

        if self.max_height and height > self.max_height:
            height = self.max_height

        if self.ratio:
            ratio_x, ratio_y = self.ratio

            if width / height > ratio_x / ratio_y:
                width = int(height * (ratio_x / ratio_y))

            else:
                height = int(width * (ratio_y / ratio_x))

        self.configure(width=width, height=height)

    def destroy(self):
        super().destroy()
        self.frame.destroy()


def make_window_draggable(window: Tk | Toplevel, dragged_function: Callable[[int, int], None] = lambda *_: None):
    window.drag_start = (None, None)

    def on_drag_start(event):
        window.drag_start = event.x, event.y

    def on_drag_motion(event):
        delta_x = event.x - window.drag_start[0]
        delta_y = event.y - window.drag_start[1]

        new_x = window.winfo_x() + delta_x
        new_y = window.winfo_y() + delta_y

        width, height = window.winfo_width(), window.winfo_height()
        s_width, s_height = window.winfo_screenwidth(), window.winfo_screenheight()

        new_x = min(max(-width // 2, new_x), s_width - width // 2)
        new_y = min(max(-height // 2, new_y), s_height - height // 2)

        window.geometry(f"+{new_x}+{new_y}")

        dragged_function(delta_x, delta_y)

    window.bind("<ButtonPress-1>", on_drag_start)
    window.bind("<B1-Motion>", on_drag_motion)


def after(func, window: Misc):
    """Wrapper"""

    def do(*args, **kwargs):
        window.after_idle(lambda: func(*args, **kwargs))

    return do


# Compatibility
NewTK = NewTk

__all__ = ["NewTk", "clear", "protect_grid_size", "Border", "ScrollingFrame", "ToolTip", "NewFrame", "get_default_font",
           "is_hovered", "validate_input", "NewFont", "NewLabel", "NewWidget", "hovered_widget",
           "generate_validate_command", "LockedFrame", "make_window_draggable", "NewTK", "after"]
